/*
 * Program to compute times of the FFT
 * Coro.
 *  modificated for BSP. euroPVM/MPI 99
 * 13-05-98.
 */



/* ----- Libraries ----- */
#include <bsp.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/time.h>
#include <stdarg.h>
#include "mytime.h"

/* ----- Constants ----- */

#define K	(1024)
#define M	(K * K)
#define MaxSIZE	 (2 * M)
#define PI       3.1415926535
#define MAX_TIMES	20
#define MAX_DEEP	20

#define PAR(f1, r1, s1, f2, r2, s2) { \
	int ll_partner, aux, status; \
	\
	NUMPROCESSORS /= 2; \
	BIT--;                            \
	ll_partner = NAME ^ (1 << BIT);                \
	if ((NAME & (1 << BIT)) == 0) {        \
		f1;                                 \
    bsp_send( ll_partner, &ll_partner, r1, s1);\
		bsp_sync(); \
		GETTIME(combinationTime1); \
		bsp_get_tag(&s2, &aux); \
		bsp_move(r2,s2); \
	}                                     \
	else {                                \
		f2;                                 \
    bsp_send( ll_partner, &ll_partner, r2, s2);\
		bsp_sync(); \
		GETTIME(combinationTime1); \
		bsp_get_tag(&s1, &aux);  \
		bsp_move(r1,s1);  \
	}                                     \
	NUMPROCESSORS *= 2;                  \
	BIT++;                               \
}

typedef float doubleType;

typedef struct {
	 doubleType re;
	 doubleType im;
} Complex;

/* ----- prototypes ----- */
void write_array(Complex *a);
void initialize(int Size, Complex *a);
void parDandCFFT(Complex *a, Complex *A,int N);
void DandCFFT(Complex *a, Complex *A, int N);

/* ----- Global variables ----- */
int NAME, NUMPROCESSORS, BIT;
Complex a[MaxSIZE], A[MaxSIZE];   /* Array para la seal y su transformada */

/* chronos */
int REP;
int deep = 0;
struct timeval time1, time2;
double t_division[MAX_TIMES];
double t_combination[MAX_TIMES][MAX_DEEP];
double t_seq[MAX_TIMES];
/* ----- Functions ----- */

/************************************************************/

void initialize (int SIZE, Complex *a) {
	unsigned i;
	for(i = 0; i < SIZE; i++) {
		a[i].re = 1.0;
		a[i].im = 0.0;
	}
}

/*
 * Write results
 */
void write_array(Complex *a) {
	unsigned i;

	for(i = 0; i < 10; i++)
		printf("a[%2u] = [%10f,%10f]\n", i, a[i].re, a[i].im);
}

/*
 * Fast Fourier Transform
 */
void DandCFFT(Complex *a, Complex *A, int N) {
	Complex *b, *B;         /* vectores para los terminos pares  de a[] */
	Complex *c, *C;         /* vectores para los terminos impares de a[] */
	int i, j, n;
	Complex wi;
	int size;

	if (N == 1) {           /* Fase de resolucion del problema trivial */
		A[0].re = a[0].re;
		A[0].im = a[0].im;
	}
	else {
		n = N / 2;
		size = n * sizeof(Complex);
		b = (Complex *)malloc(4 * size);
		if (b == NULL) {
			bsp_abort("Error inside DancCFFT allocating memory...\n");
		}
		B = b + n;
		c = B + n;
		C = c + n;
		/* Fase de division del problema en dos subproblemas de tamanno N/2 */
		for (i = 0, j = 0; i < n; i++, j = 2 * i) {
			b[i] = a[j];        /* b[] almacena las componentes pares de a[] */
			c[i] = a[j+1];      /* c[] almacena las componentes impares de a[] */
		}
		/* Fase de resolucion de los dos subproblemas */
		DandCFFT(b, B, n);
		DandCFFT(c, C, n);
		/* Fase de combinacion de las soluciones de los subproblemas
			 para obtener la solucion del problema original */
		for (i = 0; i < n; i++) {
			wi.re = cos(2.0L * PI * i / N);
			wi.im = sin(2.0L * PI * i / N);
			/* B[i] + w^i C[i]  para las componentes desde 0 hasta N/2-1 */
			A[i].re = B[i].re + (wi.re * C[i].re - wi.im * C[i].im);
			A[i].im = B[i].im + (wi.re * C[i].im + wi.im * C[i].re);
			/* B[i] - w^i C[i] para las componentes desde N/2 hasta N-1 */
			A[i + n].re = B[i].re - (wi.re * C[i].re - wi.im * C[i].im);
			A[i + n].im = B[i].im - (wi.re * C[i].im + wi.im * C[i].re);
		}
		free(b);
	} /* else */
} /* DandCFFT()... */

/*
 * Parallel Fast Fourier Transform
 */
void parDandCFFT(Complex *a, Complex *A, int N) {
	Complex *b, *B;         /* vectores para los terminos pares  de a[] */
	Complex *c, *C;         /* vectores para los terminos impares de a[] */
	int i, j, n;
	Complex wi;
	int size;
	double	t_div_1, t_div_2,
					t_PAR_1, t_PAR_2,
					t_comb_1, t_comb_2,
					t_seqIn_1, t_seqIn_2;

	struct timeval divisionTime1, divisionTime2, combinationTime1,
		combinationTime2, seqTime1, seqTime2;

	if(NUMPROCESSORS > 1) {
		if (N == 1) {           /* Fase de resolucion problema trivial */
			A[0].re = a[0].re;
			A[0].im = a[0].im;
		}
		else {
			GETTIME(divisionTime1);
			n = N / 2;
			size = n * sizeof(Complex);
			b = (Complex *)malloc(4 * size);
			if (b == NULL) {
				bsp_abort("Error allocating memory...\n");
			}
			B = b + n;
			c = B + n;
			C = c + n;

			/* Fase de division del problema en dos subproblemas de tamanno N / 2 */
			for (i = 0, j = 0; i < n; i++, j = 2 * i) {
					b[i] = a[j];   /* b[] almacena las componentes pares de a[] */
					c[i] = a[j+1]; /* c[] almacena las componentes impares de a[] */
			}
			/* Fase de resolucion de los dos subproblemas */
			PAR(parDandCFFT(b, B, n), B, size, parDandCFFT(c, C, n), C, size);

			/* Fase de combinacion de las soluciones de los subproblemas
				 para obtener la solucion del problema original */
			for (i = 0; i < n; i++) {
					wi.re = cos(2.0L * PI * i / N);
					wi.im = sin(2.0L * PI * i / N);
					 /* B[i] + w^i C[i]  para las componentes desde 0 hasta N/2-1 */
					A[i].re = B[i].re + (wi.re * C[i].re - wi.im * C[i].im);
					A[i].im = B[i].im + (wi.re * C[i].im + wi.im * C[i].re);
					/* B[i] - w^i C[i] para las componentes desde N/2 hasta N-1 */
					A[i+n].re = B[i].re - (wi.re * C[i].re - wi.im * C[i].im);
					A[i+n].im = B[i].im - (wi.re * C[i].im + wi.im * C[i].re);
			}
			free(b);
      GETTIME(combinationTime2);
			DIFTIME(combinationTime2, combinationTime1, t_combination[REP][deep]);
			deep++;
		} /* else N == 1... */
	} /* if NUMPROCESSORS ... */
  else {
				GETTIME(seqTime1);
				DandCFFT(a, A, N);
				GETTIME(seqTime2);
				DIFTIME(seqTime2,seqTime1,t_seq[REP]);
				DIFTIME(seqTime2,time1,t_division[REP]);
	}
} /* parDandCFFT()... */

/*
 * Test output
 */
void test_array(int SIZE, Complex *a)
{
	unsigned i, OK = 1;

	if (a[0].re == SIZE)
	{
		 for(i = 1; i < SIZE; i++)
			 if (a[i].re != 0.0 || a[i].im != 0)
			 {
				 OK = 0;
				 break;
			 }
	}
	else OK = 0;
	if (!OK) printf("ERROR\n");
} /* test_array */


main(int argc, char * argv[])
{
	int SIZE, i;
	int sizeInKb;
	int MAX_REP;
	double t_sequential[MAX_TIMES], t_par[MAX_TIMES];
	Complex *basura;
  int d, name;
	int tag_size;

	bsp_begin(bsp_nprocs());
	NAME = bsp_pid();
	NUMPROCESSORS = bsp_nprocs();
	BIT = log((double)NUMPROCESSORS) / log(2.0);
	tag_size = sizeof(int);
	bsp_set_tagsize(&tag_size);
	bsp_sync();

  if (argc < 3) 
		bsp_abort("need two arguments: size in Kb and MAX_REP");

	sizeInKb = atoi(argv[1]); /* warning! bsprun filename args  is the second? */
	MAX_REP = atoi(argv[2]); 
	SIZE =  K * sizeInKb; /* K * 2^6 = 2^16 = 65.536 */

	initialize((1 << MAX_REP)*SIZE, a);
	for(REP = 0; REP < MAX_REP; REP++)
	{
		/* ----- Sequential Code ----- */
		deep = 0;
		bsp_sync();
		GETTIME(time1); 
		DandCFFT(a, A, SIZE); 
		GETTIME(time2);
		DIFTIME(time2, time1, t_sequential[REP]);
		test_array(SIZE, A);
		/* ----- Parallel Code ----- */
		bsp_sync();
		GETTIME(time1);
		parDandCFFT(a, A, SIZE);
		GETTIME(time2);
		DIFTIME(time2, time1, t_par[REP]);
		test_array(SIZE, A);
    SIZE = 2*SIZE;
	} /* for */

  fflush(stdout);
	bsp_sync();
	if (NAME == 0) { 
		printf("PARALLEL FFT WRITTEN IN BSP. EUROPVM/MPI'99. RECURSIVE COMMON-COMMON VERSION\n");
		printf("SEQUENTIAL  \tREP         \tNAME        \tDEEP        \tSIZE        \tFFT(P=1)  \tPAR        \tNUMPROC    \tSPEEDUP    \tDIVISION    \tCOMBIN     \t\n");
	}
	SIZE =  K *  sizeInKb; /* K * 2^6 = 2^16 = 65.536 */
	for(REP = 0; REP < MAX_REP; REP++)  {
		for (name = 0; name <  NUMPROCESSORS; name++) {
			if (NAME == name) {
				for (d=0; d<deep; d++) {
				  printf("%12.8lf\t",t_sequential[REP]);
					printf("%12d\t%12d\t%12d\t%12d\t",REP, NAME, d, SIZE);
				  printf("%12.8lf\t%12.8lf\t", t_seq[REP], t_par[REP]);
					printf("%12d\t",NUMPROCESSORS);
					if  (t_par[REP] != (double) 0.0) printf("%12.8lf\t", (double)(t_sequential[REP]/t_par[REP]));
					else printf("warning!!!!!!\n");
				  printf("%12.8lf\t%12.8lf\t", t_division[REP], t_combination[REP][d]);
					printf("\n");
					fflush(stdout);
				} /* for */
			}
			bsp_sync();
		}
		SIZE = 2*SIZE;
		bsp_sync();
	}

	bsp_sync();
	bsp_end();
} /* main */

